from __future__ import print_function, division, absolute_import
import time
from events import Events
from websocket import ABNF
import websocket
import gevent
from gevent import monkey
from hkube_python_wrapper.util.encoding import Encoding
from hkube_python_wrapper.wrapper.messages import messages
monkey.patch_all()


class WebsocketClient:
    def __init__(self, encoding):
        self.events = Events()
        self._ws = None
        self._reconnectInterval = 0.1
        self._active = True
        self._switcher = {
            messages.incoming.initialize: self.init,
            messages.incoming.start: self.start,
            messages.incoming.stop: self.stop,
            messages.incoming.exit: self.exit,
            messages.incoming.algorithmExecutionDone: self.algorithmExecutionDone,
            messages.incoming.algorithmExecutionError: self.algorithmExecutionError,
            messages.incoming.subPipelineDone: self.subPipelineDone,
            messages.incoming.subPipelineStarted: self.subPipelineStarted,
            messages.incoming.subPipelineError: self.subPipelineError,
            messages.incoming.subPipelineStopped: self.subPipelineStopped
        }
        self._firstConnect = False
        self._encoding = Encoding(encoding)
        self._ws_opcode = ABNF.OPCODE_BINARY if self._encoding.isBinary else ABNF.OPCODE_TEXT
        print('Initialized socket with {encoding} encoding'.format(encoding=encoding))

    def init(self, data):
        self.events.on_init(data)

    def start(self, data):
        self.events.on_start(data)

    def stop(self, data):
        self.events.on_stop(data)

    def exit(self, data):
        self.events.on_exit(data)

    def algorithmExecutionDone(self, data):
        self.events.on_algorithmExecutionDone(data)

    def algorithmExecutionError(self, data):
        self.events.on_algorithmExecutionError(data)

    def subPipelineStarted(self, data):
        self.events.on_subPipelineStarted(data)

    def subPipelineDone(self, data):
        self.events.on_subPipelineDone(data)

    def subPipelineError(self, data):
        self.events.on_subPipelineError(data)

    def subPipelineStopped(self, data):
        self.events.on_subPipelineStopped(data)

    def on_message(self, message):
        decoded = self._encoding.decode(message, plain_encode=True)
        command = decoded["command"]
        data = decoded.get("data", None)
        print('got message from worker: {command}'.format(command=command))
        func = self._switcher.get(command)
        gevent.spawn(func, data)

    def on_error(self, error):
        if self._firstConnect:
            print(error)

    def on_close(self):
        self.events.on_disconnect()

    def on_open(self):
        self._firstConnect = True
        self.events.on_connection()

    def send(self, message):
        print('sending message to worker: {command}'.format(**message))
        self._ws.send(self._encoding.encode(message, plain_encode=True), opcode=self._ws_opcode)

    def startWS(self, url):
        self._ws = websocket.WebSocketApp(
            url,
            on_message=self.on_message,
            on_error=self.on_error,
            on_open=self.on_open,
            on_close=self.on_close)
        while self._active:
            try:
                self._ws.run_forever()
                time.sleep(self._reconnectInterval)
            except Exception:
                pass

    def shutDown(self):
        self._active = False




from gevent import sleep
from hkube_python_wrapper.communication.zmq.ZMQServers import ZMQServers
from hkube_python_wrapper.util.encoding import Encoding
import hkube_python_wrapper.util.object_path as objectPath
from hkube_python_wrapper.util.decorators import timing
import hkube_python_wrapper.util.type_check as typeCheck
from hkube_python_wrapper.cache.caching import CustomCache


class DataServer:

    def __init__(self, config):
        self._cache = CustomCache(config)
        self._host = config['host']
        self._port = config['port']
        self._encodingType = config['encoding']
        self._encoding = Encoding(self._encodingType)
        self._adapter = ZMQServers(self._port, self._createReply)

    def listen(self):
        print('discovery serving on {host}:{port} with {encoding} encoding'.format(
            host=self._host, port=self._port, encoding=self._encodingType))
        self._adapter.listen()

    @timing
    def _createReply(self, message):
        try:
            decoded = self._encoding.decode(message, plain_encode=True)
            tasks = decoded.get('tasks')
            taskId = decoded.get('taskId')
            datapath = decoded.get('dataPath')
            result = self.createData(taskId, tasks, datapath)

        except Exception as e:
            result = self._createError('unknown', str(e))

        return self._encoding.encode(result)

    @timing
    def createData(self, taskId, tasks, datapath):
        if(taskId is not None):
            return self._getDataByTaskId(taskId, datapath)

        errors = False
        items = []

        for task in tasks:
            result = self._getDataByTaskId(task, datapath)
            if(typeCheck.isDict(result) and 'hkube_error' in result):
                errors = True

            items.append(result)

        return dict({"items": items, "errors": errors})

    def _getDataByTaskId(self, taskId, datapath):
        result = None
        if(taskId not in self._cache):
            result = self._createError('notAvailable', 'taskId notAvailable')
        else:
            data = self._cache.get(taskId)
            result = self._createDataByPath(data, datapath)
        return result

    def _createDataByPath(self, data, datapath):
        if(datapath):
            data = objectPath.getPath(data, datapath)
            if(data == 'DEFAULT'):
                data = self._createError('noSuchDataPath', '{datapath} does not exist in data'.format(datapath=datapath))

        return data

    def setSendingState(self, taskId, data):
        self._cache.update(taskId, data)

    def _createError(self, code, message):
        return {'hkube_error': {'code': code, 'message': message}}

    def isLocal(self, host, port):
        return host == self._host and port == self._port

    def isServing(self):
        return self._adapter.isServing()

    def shutDown(self):
        while(self.isServing()):
            sleep(1)
        self._adapter.close()



from __future__ import print_function, division, absolute_import
import multiprocessing
import concurrent.futures
import hkube_python_wrapper.util.type_check as typeCheck
from hkube_python_wrapper.util.decorators import timing, trace
from hkube_python_wrapper.util.object_path import getPath, setPath
from hkube_python_wrapper.util.encoding import Encoding
from hkube_python_wrapper.storage.storage_manager import StorageManager
from hkube_python_wrapper.communication.DataRequest import DataRequest

class DataAdapter:
    def __init__(self, options, dataServer=None):
        self._dataServer = dataServer
        self._storageCache = dict()
        self._encoding = Encoding(options.storage['encoding'])
        self._storageManager = StorageManager(options.storage)
        self._requestEncoding = options.discovery['encoding']
        self._requestTimeout = options.discovery['timeout']
        self._maxWorkers = min(32, (multiprocessing.cpu_count() or 1) + 4)
        print('using {workers} for DataAdapter'.format(workers=self._maxWorkers))

    def encode(self, value):
        return self._encoding.encode(value)

    def decode(self, value):
        return self._encoding.decode(value)

    @trace()
    def getData(self, options):
        jobId = options.get('jobId')
        inputArgs = options.get('input')
        flatInput = options.get('flatInput')
        storage = options.get('storage')
        useCache = options.get('useCache')

        if (not flatInput):
            return inputArgs

        if (useCache is False):
            self._storageCache = dict()

        for k, v in flatInput.items():
            if self._isStorage(v):
                key = v[2:]
                link = storage.get(key, None)
                if (link is None):
                    raise Exception('unable to find storage key')

                if(typeCheck.isList(link)):
                    data = self.batchRequest(link, jobId)
                else:
                    data = self.tryGetDataFromPeerOrStorage(link)

                setPath(inputArgs, k, data)

        return inputArgs

    def _isStorage(self, value):
        return typeCheck.isString(value) and value.startswith('$$')

    @trace()
    def setData(self, options):
        jobId = options.get('jobId')
        taskId = options.get('taskId')
        data = options.get('data')
        result = self._storageManager.hkube.put(jobId, taskId, data)
        return result

    @timing
    def batchRequest(self, options, jobId):
        batchResponse = []
        for d in options:
            d.update({"jobId": jobId})

        with concurrent.futures.ThreadPoolExecutor(max_workers=self._maxWorkers) as executor:
            for out in executor.map(self._batchRequest, options):
                batchResponse += out

        return batchResponse

    def _batchRequest(self, options):
        batchResponse = []
        jobId = options.get('jobId')
        tasks = options.get('tasks')
        dataPath = options.get('path')
        storageInfo = options.get('storageInfo')
        if (storageInfo):
            storageResult = self._getFromCacheOrStorage(storageInfo, dataPath)
            batchResponse.append(storageResult)
            return batchResponse
        peerResponse = self._getFromPeer(options, dataPath)
        peerError = self._getPeerError(peerResponse)

        if(peerError):
            message = peerError.get('message')
            print('batch request has failed with {message}, using storage fallback'.format(message=message))
            for t in tasks:
                storageData = self._getDataForTask(jobId, t, dataPath)
                batchResponse.append(storageData)
        else:
            errors = peerResponse.get('errors')
            items = peerResponse.get('items')

            if(errors):
                for i, t in enumerate(items):
                    peerError = self._getPeerError(t)
                    if(peerError):
                        taskId = tasks[i]
                        storageData = self._getDataForTask(jobId, taskId, dataPath)
                        batchResponse.append(storageData)
                    else:
                        batchResponse.append(t)
            else:
                batchResponse += items

        return batchResponse

    def _getDataForTask(self, jobId, taskId, dataPath):
        path = self._storageManager.hkube.createPath(jobId, taskId)
        return self._getFromCacheOrStorage({'path': path}, dataPath)

    def tryGetDataFromPeerOrStorage(self, options):
        dataPath = options.get('path')
        storageInfo = options.get('storageInfo')
        discovery = options.get('discovery')
        data = None
        hasResponse = False

        if (discovery):
            data = self._getFromPeer(options, dataPath)
            peerError = self._getPeerError(data)
            hasResponse = not peerError
            data = None if peerError else data

        if (not hasResponse and storageInfo):
            data = self._getFromCacheOrStorage(storageInfo, dataPath)

        return data

    @trace(name='getFromPeer')
    @timing
    def _getFromPeer(self, options, dataPath):
        tasks = options.get('tasks')
        taskId = options.get('taskId')
        discovery = options.get('discovery')
        port = discovery.get('port')
        host = discovery.get('host')

        response = None
        if(self._dataServer and self._dataServer.isLocal(host, port)):
            response = self._dataServer.createData(taskId, tasks, dataPath)

        else:
            request = {
                'address': {
                    'port': port,
                    'host': host
                },
                'tasks': tasks,
                'taskId': taskId,
                'dataPath': dataPath,
                'encoding': self._requestEncoding,
                'timeout': self._requestTimeout
            }
            dataRequest = DataRequest(request)
            response = dataRequest.invoke()

        return response

    def _getPeerError(self, options):
        error = None
        if(typeCheck.isDict(options)):
            error = options.get('hkube_error')

        return error

    def _getFromCacheOrStorage(self, options, dataPath):
        path = options.get('path')
        data = self._getFromCache(path)
        if (data is None):
            data = self._getFromStorage(options)
            self._setToCache(path, data)

        if(dataPath):
            data = getPath(data, dataPath)

        return data

    @trace(name='getFromCache')
    @timing
    def _getFromCache(self, path):
        return self._storageCache.get(path)

    def _setToCache(self, path, data):
        self._storageCache[path] = data

    @trace(name='getFromStorage')
    @timing
    def _getFromStorage(self, options):
        response = self._storageManager.storage.get(options)
        return self._encoding.decode(response)

    def createStorageInfo(self, options):
        jobId = options.get('jobId')
        taskId = options.get('taskId')
        encodedData = options.get('encodedData')

        path = self._storageManager.hkube.createPath(jobId, taskId)
        metadata = self.createMetadata(options)

        storageInfo = {
            'storageInfo': {
                'path': path,
                'size': len(encodedData) if encodedData else 0
            },
            'metadata': metadata
        }
        return storageInfo

    def createMetadata(self, options):
        nodeName = options.get('nodeName')
        data = options.get('data')
        savePaths = options.get('savePaths', [])

        metadata = dict()
        objData = dict()
        objData[nodeName] = data
        for path in savePaths:
            try:
                value = getPath(objData, path)
                if (value != 'DEFAULT'):
                    meta = self._getMetadata(value)
                    metadata[path] = meta
            except Exception:
                pass

        return metadata

    def _getMetadata(self, value):
        if(typeCheck.isDict(value)):
            meta = {'type': 'object'}
        elif(typeCheck.isList(value)):
            meta = {'type': 'array', 'size': len(value)}
        else:
            meta = {'type': str(type(value).__name__)}
        return meta



import zmq.green as zmq
from .ZMQServer import ZMQServer
context = zmq.Context()


class ZMQServers(object):
    def __init__(self, port, replyFunc):
        self._isServing = False
        self._active = True
        self._replyFunc = replyFunc
        self._url_worker = "inproc://workers"
        self._url_client = "tcp://*:" + str(port)
        self._instances = []
        self._clients = None
        self._workers = None

    def listen(self):
        self._clients = context.socket(zmq.ROUTER)
        self._clients.bind(self._url_client)

        self._workers = context.socket(zmq.DEALER)
        self._workers.bind(self._url_worker)

        for _ in range(5):
            server = ZMQServer(context, self._replyFunc, self._url_worker)
            server.start()
            self._instances.append(server)

        zmq.device(zmq.QUEUE, self._clients, self._workers)

        self._clients.close()
        self._workers.close()
        context.term()

    def isServing(self):
        res = all(i.isServing() for i in self._instances)
        return res

    def close(self):
        for i in self._instances:
            i.stop()
            i.close()
            i.join(timeout=1)
        self._clients.close()
        self._workers.close()
        context.term()



import datetime


class CustomCache:
    def __init__(self, config):
        self._cache = dict()
        self._maxCacheSize = config.get('maxCacheSize')

    def update(self, key, value):
        if key not in self._cache and len(self._cache) >= self._maxCacheSize:
            self._remove_oldest()
        self._cache[key] = {'timestamp': datetime.datetime.now(), 'value': value}

    def __contains__(self, key):
        return key in self._cache

    def _remove_oldest(self):
        oldest = None
        for key in self._cache:
            if oldest is None:
                oldest = key
            elif self._cache[key]['timestamp'] < self._cache[oldest]['timestamp']:
                oldest = key
        self._cache.pop(oldest)

    def get(self, key):
        item = self._cache[key]
        return item.get('value')



from __future__ import print_function, division, absolute_import
import hkube_python_wrapper.util.type_check as typeCheck
from hkube_python_wrapper.wrapper.messages import messages
from .execution import Execution
from .waitFor import WaitForData


class HKubeApi:
    def __init__(self, wc, dataAdapter, storage):
        self._wc = wc
        self._dataAdapter = dataAdapter
        self._storage = storage
        self._executions = {}
        self._lastExecId = 0
        self._wc.events.on_algorithmExecutionDone += self.algorithmExecutionDone
        self._wc.events.on_algorithmExecutionError += self.algorithmExecutionDone

        self._wc.events.on_subPipelineDone += self.subPipelineDone
        self._wc.events.on_subPipelineError += self.subPipelineDone
        self._wc.events.on_subPipelineStopped += self.subPipelineDone

    def _generateExecId(self):
        self._lastExecId += 1
        return str(self._lastExecId)

    def algorithmExecutionDone(self, data):
        execId = data.get('execId')
        self._handleExecutionDone(execId, data)

    def subPipelineDone(self, data):
        subPipelineId = data.get('subPipelineId')
        self._handleExecutionDone(subPipelineId, data)

    def _handleExecutionDone(self, execId, data):
        execution = self._executions.get(execId)

        try:
            error = data.get('error')
            if(error):
                execution.waiter.set(error)

            elif(execution.includeResult):
                response = data.get('response')
                result = response
                if(typeCheck.isDict(response) and response.get('storageInfo') and self._storage == 'v2'):
                    result = self._dataAdapter.tryGetDataFromPeerOrStorage(response)
                execution.waiter.set(result)
            else:
                execution.waiter.set(None)

        except Exception as e:
            execution.waiter.set(e)
        finally:
            self._executions.pop(execId)

    def start_algorithm(self, algorithmName, input=[], includeResult=True, blocking=False):
        print('start_algorithm called with {name}'.format(name=algorithmName))
        execId = self._generateExecId()
        execution = Execution(execId, includeResult, WaitForData(True))
        self._executions[execId] = execution

        message = {
            "command": messages.outgoing.startAlgorithmExecution,
            "data": {
                "execId": execId,
                "algorithmName": algorithmName,
                "input": input,
                "includeResult": includeResult
            }
        }
        self._wc.send(message)

        if blocking:
            return execution.waiter.get()
        return execution.waiter

    def start_stored_subpipeline(self, name, flowInput={}, includeResult=True, blocking=False):
        print('start_stored_subpipeline called with {name}'.format(name=name))
        execId = self._generateExecId()
        execution = Execution(execId, includeResult, WaitForData(True))
        self._executions[execId] = execution

        message = {
            "command": messages.outgoing.startStoredSubPipeline,
            "data": {
                "subPipeline": {
                    "name": name,
                    "flowInput": flowInput
                },
                "subPipelineId": execId,
                "includeResult": includeResult
            }
        }
        self._wc.send(message)

        if blocking:
            return execution.waiter.get()
        return execution.waiter

    def start_raw_subpipeline(self, name, nodes, flowInput, options=None, webhooks=None, includeResult=True, blocking=False):
        print('start_raw_subpipeline called with {name}'.format(name=name))
        execId = self._generateExecId()
        execution = Execution(execId, includeResult, WaitForData(True))
        self._executions[execId] = execution

        message = {
            "command": messages.outgoing.startRawSubPipeline,
            "data": {
                "subPipeline": {
                    "name": name,
                    "nodes": nodes,
                    "options": options,
                    "webhooks": webhooks,
                    "flowInput": flowInput
                },
                "subPipelineId": execId,
                "includeResult": includeResult
            }
        }
        self._wc.send(message)

        if blocking:
            return execution.waiter.get()
        return execution.waiter




class BaseStorageManager(object):
    def __init__(self, adpter):
        self.adapter = adpter

    def put(self, options):
        return self.adapter.put(options)

    def get(self, options):
        data = self.adapter.get(options)
        if(data is None):
            return None
        return data

    def list(self, options):
        return self.adapter.list(options)

    def listPrefix(self, options):
        return self.adapter.listPrefix(options)

    def delete(self, options):
        self.adapter.delete(options)



import threading
import zmq.green as zmq


class ZMQServer(threading.Thread):
    def __init__(self, context, replyFunc, workerUrl):
        self._isServing = False
        self._active = True
        self._socket = None
        self._replyFunc = replyFunc
        self._workerUrl = workerUrl
        self._context = context
        threading.Thread.__init__(self)

    def run(self):
        self._socket = self._context.socket(zmq.REP)
        self._socket.connect(self._workerUrl)

        while self._active:
            try:
                message = self._socket.recv()
                self._isServing = True
                if(message == b'Are you there'):
                    self._socket.send(b'Yes')
                else:
                    self._send(message)
                self._isServing = False
            except Exception:
                print('socket closed')

    def _send(self, message):
        toBeSent = self._replyFunc(message)
        self._socket.send(toBeSent, copy=False)

    def isServing(self):
        return self._isServing

    def stop(self):
        self._active = False

    def close(self):
        self._socket.close()
